#include <stdio.h>
#include <ctype.h>
#include "pipe.h"

#define MAX_STAGE 10

void *malloc();

static stage_ptr stages[MAX_STAGE];
static int stage_count = 0;

stage_ptr new_stage(int count)
{
  stage_ptr result = (stage_ptr) malloc(sizeof(stage_ele));
  result->current = malloc(count);
  result->next = malloc(count);
  bzero(result->current, count);
  bzero(result->next, count);
  result->count = count;
  result->bubble = 0;
  result->stall = 0;
  stages[stage_count++] = result; 
  return result;
}

void update_stages()
{
  int s;
  for (s = 0; s < stage_count; s++) {
    stage_ptr p = stages[s];
    if (p->bubble)
      bzero(p->current, p->count);
    else if (!p->stall)
      bcopy(p->next, p->current, p->count);
    p->bubble = 0;
    p->stall = 0;
  }
}

/******************** Utility Code *************************/

/* Representations of digits */
static char digits[16] =
   {'0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

/* Print hex/oct/binary format with leading zeros */
/* bpd denotes bits per digit  Should be in range 1-4,
   pbw denotes bits per word.*/
void wprint(unsigned x, int bpd, int bpw, FILE *fp)
{
  int digit;
  unsigned mask = (1 << bpd) - 1;
  for (digit = (bpw-1)/bpd; digit >= 0; digit--) {
    unsigned val = (x >> (digit * bpd)) & mask;
    putc(digits[val], fp);
  }
}


/* Dump nonzero state of a memory, where
   state is an array of unsigned's denoting the memory state,
   count is the number of words.
   if text nonnull, prints associated text with each word
   Otherwise, puts in multicolumn format for legibility
*/
void dump_state(unsigned *state, char **text, int count, FILE *fp)
{
  int col = 0;
  int col_limit = text ? 1 : 4; 
  int bpa = 0;
  int i;
  /* figure out how many bits it takes to represent the address */ 
  while (1<<bpa < count)
    bpa++;
  for (i = 0; i < count; i++) {
    unsigned val = state[i];
    char *info = text ? text[i] : NULL;
    if (val != 0 || info != NULL) {
      fputs("  ", fp);
      wprint(i, 4, bpa, fp);
      fputs(":0x", fp);
      wprint(val, 4, 32, fp);
      if (info)
	fprintf(fp, "\t%s", info);
      if (++col >= col_limit) {
	col = 0;
	fputs("\n", fp);
      }
    }
  }
  if (col != 0)
    fputs("\n", fp);
}

/* Get string starting with first nonwhite character and ending with newline */
static char *get_text(FILE *fp)
{
  char buffer[256];
  char *result;
  int i = 0;
  char  c = getc(fp);  
  while (c != EOF && c != '\n' && isspace(c))
    c = getc(fp);
  while (c != EOF && c != '\n') {
    buffer[i++] = c;
    c = getc(fp);
  }
  buffer[i++] = '\0';
  result = malloc(i);
  strcpy(result, buffer);
  return result;
}

/* Load state from memory */
/* Assumes stored as set of pairs AAA:0xDDDD, all in Hex format */
/* If text nonnull, parses quoted strings following numeric values
   as associated text information */
/* Returns number of words read */

unsigned load_state(unsigned *state, char **text, FILE *fp)
{
  unsigned a, d;
  unsigned count = 0;
  while (fscanf(fp, "%x:0x%x", &a, &d) == 2) {
    state[a] = d;
    if (text)
      text[a] = get_text(fp);
/* DEBUG
   printf("Read word %x:0x%x\t%s\n", a, state[a], text ? text[a] : "");
*/
    count++;
  }
  return count;
}

unsigned clear_state(unsigned *state, int count)
{
  bzero(state, count*4);
}
